package com.urbancode.anthill3.plugin.clearcasebasedynamic

import com.urbancode.anthill3.AHPTool
import com.urbancode.anthill3.CommandHelper
import com.urbancode.anthill3.DateParser

import java.text.SimpleDateFormat
import java.util.List;

/* example changelog output
Action:checkin
Date:20070816.205608
User:svc-bld
Object:/vobs/dsm_proto///build.xml@@/main/dsm_eolian-dev/40
Labels:
Event:create version
Comment:
----->
Action:checkin
Date:20070816.205610
User:svc-ble
Object:/vobs/dsm_proto///build.xml@@/main/dsm_eolian-dev/41
Labels:label
Event:create version
Comment:kensie
----->
*/

public class ClearCaseHelper {

    /*--------------------------------------------------------------------------------
     *  Class
     --------------------------------------------------------------------------------*/
    static final def GMT = TimeZone.getTimeZone("GMT")
    static final def CC_DATE_FORMAT = new SimpleDateFormat("dd-MMM-yy.HH:mm:ss")
    static final def AHP_DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S Z")
    static {
        CC_DATE_FORMAT.timeZone = GMT
        AHP_DATE_FORMAT.timeZone = GMT
    }

    static final def dateParser = new DateParser()

    /**
    * Converts an Ant-style wild-mat string into an equivalent java.util.regex.Pattern
    *   ? is a single char wildcard
    *   * is a wildcard matches any string (but does not cross directories)
    *   ** is a wildcard which matches any string and can cross any number of directories
    */
    static public def convertToPattern( def antWildMat) {
       // trim leading / character from pattern
       def pattern = antWildMat.startsWith('/') ? antWildMat : '/'+antWildMat

       // deal with special regex-characters that should be interpreted as literals
       '\\.+[]^${}|()'.toCharArray().each{ c ->
           pattern = pattern.replace(''+c, '\\'+c)
       }
       pattern = pattern.replace('?', '.') // ? is a single-char wildcard

       // deal with ant-style wildcards
       StringBuffer result = new StringBuffer()
       def m = (pattern =~ '\\*\\*/|\\*\\*|\\*')
       while (m) {
           def token = m.group()
           def replacement
           if (token == '**/') {
               replacement = '.*(?<=/)'
           }
           else if (token == '**') {
               replacement = '.*'
           }
           else {
               replacement = '[^/]*'
           }
           m.appendReplacement(result, java.util.regex.Matcher.quoteReplacement(replacement))
       }
       m.appendTail(result)
       return ~(result.toString())
    }

    /*--------------------------------------------------------------------------------
     * Instance
     --------------------------------------------------------------------------------*/
    def commandPath
    def viewPath
    File workDirFile = null
    Date startDate = null
    Date endDate = null
    def vobTagRoot
    def loadRuleList = new ArrayList<String[]>()
    def ch
    def userExcludes
    def fileFilters
    def label
    def message
    def globalLabels
    def labelVobNames = []
    def includeFiles = []
    def excludeFiles = []
    def ahpTool = new AHPTool()
    def changeSets = [] //list of arrays [id,user,message,date,fileSet[]]
    def useNco
    def qpUseNco

    public ClearCaseHelper(Map<String,String> props) {
        commandPath = props['source/repo/commandPath']
        viewPath = props['source/storageDir']
        workDirFile = new File(viewPath)

        if (props['startDate'] != null) {
            println "StartDate = " + props['startDate']
            startDate = dateParser.parseDate(props['startDate'])
        }

        if (props['endDate'] != null) {
            println "EndDate = " + props['endDate']
            endDate = dateParser.parseDate(props['endDate'])
        }

        vobTagRoot = props['source/tagRoot']?:""

        //load rules
        String[] loadRuleTemp = props['source/loadRules'].split('\n')
        loadRuleList = new ArrayList<String[]>()

        loadRuleTemp.each { loadRule ->
            int i=0
            String [] tempLoadRuleArray = loadRule.split(':')
            String [] loadRuleArray = new String[4]
            for (i=0; i<tempLoadRuleArray.length && i < 4; i++) {
              loadRuleArray[i] = tempLoadRuleArray[i]
            }
            for (i; i<4; i++ ) {
              loadRuleArray[i] = null
            }
            loadRuleList << loadRuleArray
        }

        //view path
        if (viewPath == null || viewPath.trim().length() == 0) {
            throw new IllegalStateException("View path does is not configured.")
        }

        //commandpath
        if (commandPath == null || commandPath.trim().length() == 0) {
            commandPath = "cleartool"
        }


        ch = new CommandHelper(workDirFile)

        userExcludes = props['source/excludeUsers']
        if (userExcludes) {
          userExcludes = userExclude.readLines().collect{it.trim()}.findAll{it}
        }

        //--------------------------------------------------------------------------------
        //file excludes
        //--------------------------------------------------------------------------------
        fileFilters = props['fileFilters']
        if (fileFilters) {
          fileFilters.readLines().collect{it.trim()}.findAll{it}.each{filter ->
              def pattern = convertToPattern(filter.substring(1))
              if (filter.startsWith('+')) {
                  includeFiles << pattern
              }
              else if (filter.startsWith('-')) {
                  excludeFiles << pattern
              }
              else {
                  // not a valid expression
              }
          }
        }

        label = props['label']
        message = props['message']
        labelVobNames = props['labelVobs']?.split('\n')
        globalLabels = props['globalLabels']
        useNco = props['useNco']
        qpUseNco = props['source/useNco']
    }

    //--------------------------------------------------------------------------------
    //determine if it is an acceptable revision closure not testing user or file excludes
    //--------------------------------------------------------------------------------

    public boolean isRevisionAcceptable(def changeSet) {
        boolean result = false
        String fileName = changeSet['file']
        String labels = changeSet['label']
        String event = changeSet['event']
        Date revisionDate = changeSet['date']
        String action = changeSet['action']


        // find the rule for this revision
        println "Checking if revision is acceptable"
        if (changeSet['file'] != null) {
            int i = 0
            loadRuleList.each { loadRule ->
                // check and make sure the revision is in the date interval we need
                if (endDate == null || (revisionDate == null || endDate.after(revisionDate))) {
                    // now make sure that revision belongs to one of the paths specified in
                    // the list of changelog rules
                    if (fileName.indexOf(loadRule[0]) != -1 && fileName.indexOf(loadRule[1]) != -1) {
                        // make sure that the revision branch matches the changelog branch for that path
                        if (loadRule[2] == null ||
                                fileName.indexOf(loadRule[2]) != -1) {
                            // check if we have a mkbranch action that does not affect code
                            if ("mkbranch".equals(action) && loadRule[2] != null &&
                                    (fileName().endsWith(loadRule[2]) ||
                                        fileName().endsWith(loadRule[2] + "\\0") ||
                                        fileName().endsWith(loadRule[2] + "/0") )) {
                                    println "Found a maintenance mkbranch action"
                            }
                            // make sure that label specified in the changelog rule
                            // is either in the labels or event field of the revision
                            // and it needs to be in the event field for mklabel and rmlabel events
                            else if (loadRule[3] == null) {
                                println "No label specified"
                                result = true
                            }
                            else if ("mklabel".equals(action) || "rmlabel".equals(action)) {
                                println "Label was made or removed"
                                if (event != null && (event.indexOf("\"" + loadRule[3] + "\"") != -1)) {
                                    println "A matching label was made or removed"
                                    result = true
                                }
                                else {
                                    println "Label did not match the changelog label for this path"
                                    println "loadRule[3]:" + loadRule[3]
                                }
                            }
                            else if (labels != null && labels.indexOf(loadRule[3]) != -1) {
                                println "Found a change that matched our label"
                                result = true
                            }
                            else {
                                println "A label existed but did not match our load rules"
                                println "loadRule[3]:" + loadRule[3]
                                if (labels != null) {
                                    println "labels:" + labels
                                }
                            }
                        }
                        else {
                            println "revision branch does not match the changelog branch for this path"
                            println "loadRule[2]:" + loadRule[2]
                        }
                    }
                    else {
                        println "revision does not belong to one of the paths specified in the list of changelog rules"
                        println "fileName:" + fileName
                        println "loadRule[0]:" + loadRule[0]
                        println "loadRule[1]:" + loadRule[1]
                    }
                }
                else {
                    println "Revision is not in our interval"
                    if (endDate != null) {
                        println "endDate:" + endDate
                    }
                    if (revisionDate != null) {
                        println "revisionDate:" + revisionDate
                    }
                }
            }
        }
        else {
            println "File name is null"
        }

        println "is revision acceptable " + result
        return result
    }

    //--------------------------------------------------------------------------------
    //determine if it is an acceptable user
    //--------------------------------------------------------------------------------

    public boolean isUserAcceptable(def changeSet) {
        boolean result = true
        String fileName = changeSet['file']
        String labels = changeSet['label']
        String event = changeSet['event']
        Date revisionDate = changeSet['date']
        String action = changeSet['action']
        String user = changeSet['user']

        if (userExcludes) {
            result = !userExcludes.contains(user)
        }
        println "Is user accetpable : " + result

        return result
    }

    //--------------------------------------------------------------------------------
    //determine if it is an acceptable user
    //--------------------------------------------------------------------------------

    public boolean isFileAcceptable(def changeSet) {
        boolean result = true
        String fileName = changeSet['file']
        String labels = changeSet['label']
        String event = changeSet['event']
        Date revisionDate = changeSet['date']
        String action = changeSet['action']

        if (excludeFiles) {
            result = excludeFiles.find{fileName ==~ it} != null
        }

        println "Is file accetpable : " + result
        return result
    }

    //--------------------------------------------------------------------------------
    // get the changesets from the output
    //--------------------------------------------------------------------------------
    public void parseOutputStream(def out) {

        //first we should split on --------->\n<---------
        println "Output"
        println("--------------------------------------------------------------------")
        println out.toString()
        println("--------------------------------------------------------------------")
        def revsTemp = out.toString().split("--urbancodesplit-->")
        def revs = []
        //remove trailing newline
        for (int j = 0; j < revsTemp.length-1; j++) {
            revs << revsTemp[j]
        }
        revs.each { revString ->
            //next we should split on \n
            def rev = [:]

            if (revString.split("\n").length < 7) {
                //we have to have 7 lines here to get any useful information out of a revision
                //if not we should skip it
                println "Skipping because not enough fields!"
            }
            else {
                def params = revString.split("\n")
                int i=0
                if (!params[i].startsWith("Action")) {
                    //skip it
                    i++
                }
                rev['action'] = params[i].substring(params[i].indexOf(':')+1)
                i++
                rev['date'] = dateParser.parseDate(params[i].substring(params[i].indexOf(':')+1))
                i++
                rev['user'] = params[i].substring(params[i].indexOf(':')+1)
                i++
                rev['file'] = params[i].substring(params[i].indexOf(':')+1)
                i++
                rev['label'] = params[i].substring(params[i].indexOf(':')+1)
                i++
                rev['event'] = params[i].substring(params[i].indexOf(':')+1)
                i++
                rev['comment'] = params[i].substring(params[i].indexOf(':')+1)
                i++
                while (i<params.length) {
                    //these must all be comments
                    rev['comment'] = rev['comment'] + '\n' + params[i]
                }

                //only add to the rev list if we determine it is acceptable
                if (isRevisionAcceptable(rev) && isUserAcceptable(rev) && isFileAcceptable(rev)) {
                    changeSets.add(rev)
                }
            }
        }
    }

    //--------------------------------------------------------------------------------
    //Upload the changesets to anthill
    //--------------------------------------------------------------------------------
    public void uploadChangeSetsToAnthill() {
        //upload the result to AHPTool
        if (changeSets) {
            def xml = groovy.xml.DOMBuilder.newInstance().'change-log'{
                changeSets.each{ changeSet ->
                    'change-set'{
                        //id(changeSet.id)
                        'user'(changeSet['user'])
                        'date'(AHP_DATE_FORMAT.format(date))
                        if (changeSet['comment']) {
                            'comment'(changeSet['comment'])
                        }
                        if (changeSet['label']) {
                            'module'(changeSet['label'])
                        }
                        'repository-type'('ClearCase')
                        'repository-id'(System.getenv()['AH_REPOSITORY_ID'])
                        'file-set'{
                            'file'('change-type':changeSet['event'], changeSet['file'])
                        }
                    }
                }
            }

            ahpTool.addChangeSets(xml)
        }
        else {
            println "No changes detected"
        }
    }

    //--------------------------------------------------------------------------------
    //Execute the changelog function for clearcase for each loadRule
    //--------------------------------------------------------------------------------
    public void getChangeLogStep(def isQp) {
        loadRuleList.each { loadRule ->
            String fileName = loadRule[1] ?: '.'

            if (ahpTool.isWindows && vobTagRoot) {
                fileName = vobTagRoot + File.separator + fileName
            }

            String branchName = loadRule[2]
            String labelName = loadRule[3]

            println("Getting changelog for " + fileName + " in working directory " +
                workDirFile.getCanonicalPath())
            File file = new File(workDirFile, fileName)
            ByteArrayOutputStream out = new ByteArrayOutputStream()

            if (file.exists()) {
                // get the lshistory just for the directory
                def cmdArgs = [commandPath, 'lshistory']

                if (useNco || (isQp && qpUseNco)) {
                    cmdArgs << '-nco'
                }

                cmdArgs << '-r'

                if (labelName) {
                    cmdArgs << labelName
                }

                cmdArgs << '-branch'

                if (branchName) {
                    cmdArgs << branchName
                }
                else {
                    cmdArgs << 'main'
                }

                cmdArgs << '-since'
                cmdArgs << CC_DATE_FORMAT.format(startDate)+"UTC+00:00"
                cmdArgs << '-fmt'
                cmdArgs << "Action:%o\\nDate:%Nd\\nUser:%u\\nObject:%n\\nLabels:%l\\nEvent:%e\\nComment:%Nc\\n--urbancodesplit-->\\n"
                cmdArgs << fileName

                ch.runCommand("Getting History for files and subdirectory in directory.", cmdArgs) { proc ->
                    proc.out.close()
                    proc.waitForProcessOutput(out, System.out)
                }

                //parse the output adding changesets as appropriate to the changeSets array
                parseOutputStream(out)
            }
            else {
                println "Did not find file for changelog at " + file.getAbsolutePath()
            }
        }
    }

    //--------------------------------------------------------------------------------
    // Upload last updated time to the server
    //--------------------------------------------------------------------------------
    public void uploadLastUpdatedTime() {
        if (!changeSets) {
            println "No changes to use for last time!"
        }
        else {
            Date lastUpdatedTime = null
            changeSets.each { changeSet ->
                if (lastUpdatedTime == null || changeSet['date'].compareTo(lastUpdatedTime) > 0) {
                    lastUpdatedTime = changeSet['date']
                }
            }
            ahpTool.setQuietPeriodLatestTime(lastUpdatedTime.getTime().longValue())
        }
    }

    //--------------------------------------------------------------------------------
    // Create label type
    //--------------------------------------------------------------------------------
    public void makeLabelTypeCommand(String vobName) {

        def cmdArgs = [commandPath, 'mklbtype']

        if (message == null) {
            cmdArgs << "-nc"
        }
        else {
            cmdArgs << "-c"
            cmdArgs << message
        }

        if (globalLabels) {
            cmdArgs << "-global"
            cmdArgs << "-acquire"
        }
        else {
            cmdArgs << "-ordinary"
        }

        cmdArgs << label + "@" + vobName

        ch.runCommand("Creating Label Type!", cmdArgs)
    }

    //--------------------------------------------------------------------------------
    //  Get the vob name list for labeling
    //--------------------------------------------------------------------------------
    public List getVobNameList() {
        List vobNameList = new ArrayList()
        if (labelVobNames == null || labelVobNames.length == 0) {
            loadRuleList.each { loadRule ->
                String vobName = getVobFromPath(loadRule[0])
                if (vobName != null && vobName.trim().length() > 0 && !vobNameList.contains(vobName.trim())) {
                    vobNameList.add(vobName.trim())
                }
            }
        }
        else {
            labelVobNames.each { labelVobName ->
                if (labelVobName != null && labelVobName.trim().length() > 0 && !vobNameList.contains(labelVobName.trim())) {
                    vobNameList.add(labelVobName.trim())
                }
            }
        }
        return vobNameList
    }

    //--------------------------------------------------------------------------------
    //  Get the vob from path
    //--------------------------------------------------------------------------------
    private String getVobFromPath(String path) {
        String result = null

         // strip any / or \ at the beginning of the path
        while (path.startsWith("\\") || path.startsWith("/")) {
            path = path.substring(1)
        }
        if (path != null && path.length() > 0) {
            // now look for the first / or \ and strip anything after to get the name
            // of the vob.
            int index = path.indexOf('\\')
            if (index != -1) {
                result = path.substring(0, index)
            }
            else if ((index = path.indexOf('/')) != -1) {
                result = path.substring(0, index)
            }
            else {
                result = path
            }
        }
        return result
    }

    //--------------------------------------------------------------------------------
    //  Apply Labels
    //--------------------------------------------------------------------------------
    public void applyLabels(List vobNameList) {
        loadRuleList.each { loadRule ->
            String vobName = loadRule[0]
            String dirName = loadRule[1]
            if (vobName != null && vobName.trim().length() > 0) {
                println "Trying to label " + dirName
                makeLabelCommand(dirName.trim(), true)
                //next the label needs to be applied to the vob
                applyLabelToVob(vobName)
            }
        }
    }

    //--------------------------------------------------------------------------------
    //  apply label command
    //--------------------------------------------------------------------------------
    void makeLabelCommand(String dirToLabel, boolean isRecurse) {

        def cmdArgs = [commandPath, 'mklabel', '-replace']


        if (isRecurse) {
            cmdArgs << '-recurse'
        }

        if (message == null) {
            cmdArgs << '-nc'
        }
        else {
            cmdArgs << '-c'
            cmdArgs << message
        }

        cmdArgs << label
        cmdArgs << dirToLabel

        ch.runCommand("Applying label to directory " + dirToLabel, cmdArgs)
    }

    //--------------------------------------------------------------------------------
    //  apply label to vob
    //--------------------------------------------------------------------------------
    void applyLabelToVob(String vobName) {

        def cmdArgs = [commandPath, 'mklabel', '-replace']

        if (message == null) {
            cmdArgs << '-nc'
        }
        else {
            cmdArgs << '-c'
            cmdArgs << message
        }
        cmdArgs << label
        cmdArgs << vobName
        ch.runCommand("Applying label to vob " + vobName, cmdArgs)
    }
}